#include "isr_compat.h"
#include <msp430.h>
#include <stdint.h>
#include "bits.h"

// {{{ IO pins

// PORT2
#define MOTION_REG(REG) (P2 ## REG)
const uint8_t PIN_MOTION1 = BIT6;
const uint8_t PIN_MOTION2 = BIT7;

// PORT1
#define BUTTON_REG(REG) (P1 ## REG)
const uint8_t PIN_BUTTON = BIT3;

#define LED_REG(REG) (P1 ## REG)
const uint8_t PIN_LED1 = BIT0;
const uint8_t PIN_LED2 = BIT6;

// inversed - lights, normal - Launchpad
const bool LEDS_INVERSED = true;

// }}}


// {{{ Flash modes

// 0 - off
// 1 - on
// f - flash - turn on for about 7 ms, then turn off
struct Mode {
  const char* pattern[2];
};

const Mode g_modes[4] = {
  {"00000000", "00000000"},
  {"11111111", "11111111"},
  {"11001100", "00110011"},
  {"f0f1010f", "10ffff00"},
};

// }}}

// {{{ time constants

// }}}

// {{{ State

volatile bool s_button_pressed = false;
volatile bool s_motion_detected = false;

// }}}

// {{{ Motion sensor
namespace motion {
  bool getSensor1() {
    return !getBit(MOTION_REG(IN), PIN_MOTION1);
  }

  bool getSensor2() {
    return !getBit(MOTION_REG(IN), PIN_MOTION2);
  }

  void disablePullup() {
    MOTION_REG(OUT) = 0;
  }

  void enablePullup() {
    MOTION_REG(OUT) = PIN_MOTION1 + PIN_MOTION2;
    // turn active sensor off
    MOTION_REG(OUT) = MOTION_REG(IN);
  }

  void enableInterrupt(bool enable) {
    if (enable) {
      MOTION_REG(IE) = PIN_MOTION1 + PIN_MOTION2;
      enablePullup();
    } else {
      MOTION_REG(IE) = 0;
      disablePullup();
    }
  }

  void onInterrupt(uint8_t ifg) {
    if(getBits(ifg, PIN_MOTION1 + PIN_MOTION2 )) {
      s_motion_detected = true;
      // wait for the other sensor interrupt
      enablePullup();
      LPM4_EXIT;
    }
  }

  void setup() {
    // primary function
    MOTION_REG(SEL) = 0;

    // input, pull-up resistor enabled
    // MOTION_REG(DIR) = 0; // default
    MOTION_REG(REN) = PIN_MOTION1 + PIN_MOTION2;

    // high-to-low trasition
    MOTION_REG(IES) = PIN_MOTION1 + PIN_MOTION2;
  }
}

// }}}

// {{{ LEDs
namespace led {
  void setup() {
    setBits(LED_REG(DIR), PIN_LED1 + PIN_LED2);
  }

  void enableLed1(bool enable) {
    if (LEDS_INVERSED) {
      enableBits(LED_REG(OUT), PIN_LED1, !enable);
    } else {
      enableBits(LED_REG(OUT), PIN_LED1, enable);
    }
  }

  void enableLed2(bool enable) {
    if (LEDS_INVERSED) {
      enableBits(LED_REG(OUT), PIN_LED2, !enable);
    } else {
      enableBits(LED_REG(OUT), PIN_LED2, enable);
    }
  }

  const uint8_t g_led_values[4] = {
    LEDS_INVERSED ? PIN_LED1 + PIN_LED2 : 0,
    LEDS_INVERSED ?            PIN_LED2 : PIN_LED1,
    LEDS_INVERSED ? PIN_LED1            :            PIN_LED2,
    LEDS_INVERSED ?                   0 : PIN_LED1 + PIN_LED2};

  void setLeds(uint8_t ledMask) {
    uint8_t value = g_led_values[ledMask];
    setBitsTo(LED_REG(OUT), PIN_LED1 + PIN_LED2, value);
  }

  void off() {
    setBitsTo(LED_REG(OUT), PIN_LED1 + PIN_LED2, g_led_values[0]);
  }
}

// }}} LEDs

// {{{ Button

namespace button {
  void enableInterrupt(bool enable) {
    enableBits(BUTTON_REG(IE), PIN_BUTTON, enable);
  }

  bool pressed() {
    return getBits(BUTTON_REG(IN), PIN_BUTTON) == 0;
  }

  void setup() {
    // primary function
    //unsetBits(BUTTON_REG(SEL), PIN_BUTTON); // default

    // input, pull-up resistor enabled
    //unsetBits(BUTTON_REG(DIR), PIN_BUTTON); // default
    setBits(BUTTON_REG(REN), PIN_BUTTON);
    setBits(BUTTON_REG(OUT), PIN_BUTTON);

    // high-to-low trasition
    setBits(BUTTON_REG(IES), PIN_BUTTON);
    enableInterrupt(true);
  }

  void onInterrupt(uint8_t ifg) {
    if(getBit(ifg, PIN_BUTTON)) {
      s_button_pressed = true;
      LPM4_EXIT;
    }
  }
}

// }}}

// {{{ Timer

namespace timer {
  void enable(bool enabled) {
    if (enabled) {
      TACTL |= TAIE;
    } else {
      TACTL &= ~TAIE;
    }
  }

  void setup() {
    TACCR0 = 61; // (12kHZ / 8 = 1500 Hz)  1500 / 75 -> 20Hz
    TACTL = TASSEL_1 // ACLK
          + ID_0     // divider 1
          + MC_1     // up to TACCR0
          + TACLR    // clear timer state
          + TAIE;    // enable interrupt
  }

  void onInterrupt() {
    LPM3_EXIT;
  }
}

// }}}

// {{{ Interrupts

// {{{ Timer
ISR(TIMER0_A1, timer_interrupt) {
  if(TACTL & TAIFG) {
    timer::onInterrupt();
    TACTL = TACTL & ~TAIFG;
  }
}

// }}} Timer

// {{{ IO Ports

ISR(PORT1, port1_interrupt) {
  button::onInterrupt(P1IFG);
  P1IFG = 0;
}

ISR(PORT2, port2_interrupt) {
  motion::onInterrupt(P2IFG);
  P2IFG = 0;
}

// }}} IO ports

// }}} Interrupts


const uint16_t MAX_CYCLES = (10 * 60 * 20 / 8 + 1) * 8; // ~10 min, rounded to 8
//const uint16_t MAX_CYCLES = 5 * 20; // 5.8 sec
uint16_t cycles_count = 0;

enum State {
  STATE_NORMAL = 0,
  STATE_WAIT_FOR_SETTLE = 2,
  STATE_SLEEP = 4,
};

#define MKPATTERN(c0,c1,c2,c3,c4,c5,c6,c7) \
  (((uint16_t)(c7 == '1') << 15) + (uint16_t)((c7 != '0') << 14) + \
   ((uint16_t)(c6 == '1') << 13) + (uint16_t)((c6 != '0') << 12) + \
   ((uint16_t)(c5 == '1') << 11) + (uint16_t)((c5 != '0') << 10) + \
   ((uint16_t)(c4 == '1') <<  9) + (uint16_t)((c4 != '0') <<  8) + \
   ((uint16_t)(c3 == '1') <<  7) + (uint16_t)((c3 != '0') <<  6) + \
   ((uint16_t)(c2 == '1') <<  5) + (uint16_t)((c2 != '0') <<  4) + \
   ((uint16_t)(c1 == '1') <<  3) + (uint16_t)((c1 != '0') <<  2) + \
   ((uint16_t)(c0 == '1') <<  1) + (uint16_t)((c0 != '0') <<  0))

const uint16_t light1_pattern[4] = {
  0,
  0xffffu,
  MKPATTERN('1', '1', '0', '0', '1', '1', '0', '0'),
  MKPATTERN('f', '0', 'f', '1', '0', '1', '0', 'f'),
};
const uint16_t light2_pattern[4] = {
  0,
  0xffffu,
  MKPATTERN('0', '0', '1', '1', '0', '0', '1', '1'),
  MKPATTERN('1', '0', 'f', 'f', 'f', 'f', '0', '0'),
};

uint16_t light1_tmp;
uint16_t light2_tmp;

State state = STATE_NORMAL;
const uint8_t MODE_SLEEP = 0;
uint8_t mode = MODE_SLEEP;

void delay_cycles(uint16_t cycles) {
  // 5 cycles per iteration
  for (uint16_t i = 0; i < cycles / 5; i++) {
    __nop();
  }
}

// ~20Hz loop
void loop()
{
  switch (state) {
    default:
    case STATE_NORMAL:
      if (s_button_pressed) {
        mode = (mode + 1) & 0x03;
        state = STATE_WAIT_FOR_SETTLE;
        cycles_count = 0;
        motion::enableInterrupt(mode != MODE_SLEEP);
      }

      if ((cycles_count & 0x07) == 0) {
        light1_tmp = light1_pattern[mode];
        light2_tmp = light2_pattern[mode];

        // handle motion sensor event only when blink step is the same
        // this is needed to reset counter without having "stale" leds
        if (s_motion_detected) {
          s_motion_detected = false;
          cycles_count = 0;
        }
      }

      {
        uint8_t carryouts = shiftRightAndReturnCarry(light1_tmp, light2_tmp);
        led::setLeds(carryouts);
      }

      cycles_count++;

      {
        uint8_t carryouts = shiftRightAndReturnCarry(light1_tmp, light2_tmp);
        led::setLeds(carryouts);
      }

      if (cycles_count == MAX_CYCLES) {
        state = STATE_SLEEP;
        cycles_count = 0;
      }

      break;

    case STATE_WAIT_FOR_SETTLE:
      {
        // wait for 1s to detect long press
        // CPU runs at 12kHz
        delay_cycles(12000);

        state = STATE_NORMAL;

        // still pressed after pause?
        if (button::pressed()) {
          // long press detected
          state = STATE_SLEEP;
          mode = MODE_SLEEP;
          motion::enableInterrupt(false);
        }

        s_button_pressed = false;
      }
      break;

    case STATE_SLEEP:
      // prepare to sleep
      led::off();

      // LPM4 disables clocks => timer is disabled too
      LPM4;
      // something happened - either motion or button event

      if (mode == MODE_SLEEP) {
        // ... that was a button, for sure
        // transit to NORMAL, it will consume button event and switch to next mode
        state = STATE_NORMAL;
      } else {
        // SETTLE will consume button event
        state = STATE_WAIT_FOR_SETTLE;
      }
      break;
  }

  LPM3;
}

void main_program() {
  WDTCTL = WDTPW + WDTHOLD;

  // Clock
  BCSCTL1 = XT2OFF + DIVA_3; // no XT2, ACLK = VLO / 8 = 1500 Hz
  BCSCTL2 = SELM_3  // LFXT1CLK or VLOCLK
          + SELS    // LFXT1CLK or VLOCLK when XT2 oscillator not present
          + DIVM_3; // MCLK /= 8
  BCSCTL3 = LFXT1S_2; // VLOCLK

  button::setup();
  led::setup();
  motion::setup();
  timer::setup(); // also enables timer

  __eint();

  // initial state is NORMAL
  // trigger switch to mode 1 and setting up inputs
  s_button_pressed = true;

  while (true) {
    loop();
  }
}


// {{{ Linker constants

// raw values from linker
extern "C" void *__stack;
extern "C" void *__bssstart;
extern "C" void *__bsssize;

// typed constants

uint16_t * const stack = reinterpret_cast<uint16_t*>(&__stack);
uint8_t * const bssstart = reinterpret_cast<uint8_t*>(&__bssstart);
uint16_t * const bssstart_w = reinterpret_cast<uint16_t*>(bssstart);
const uint16_t bsssize = reinterpret_cast<uint16_t>(&__bsssize);

// }}} Linker constants

extern "C"
__attribute__((naked))
ISR(RESET, reset_interrupt) {
  _set_SP_register(stack);

  // clear bss memory, 16bit word at a time
  for (uint16_t i = 0; i <= bsssize; i += 2) {
    bssstart_w[i] = 0;
  }

  main_program();
}
